from __future__ import absolute_import, division, print_function

from ten99policy.api_resources.abstract.api_resource import APIResource


class SingletonAPIResource(APIResource):
    @classmethod
    def retrieve(cls, **params):
        return super(SingletonAPIResource, cls).retrieve(None, **params)

    @classmethod
    def class_url(cls):
        if cls == SingletonAPIResource:
            raise NotImplementedError(
                "SingletonAPIResource is an abstract class.  You should "
                "perform actions on its subclasses (e.g. Balance)"
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        return "/api/v1/%s" % (base,)

    def instance_url(self):
        return self.class_url()



import unittest
from ten99policy.ten99policy_response import (
    Ten99PolicyResponse,
    Ten99PolicyStreamResponse,
    Ten99PolicyResponseBase,
)
import json
from collections import OrderedDict


class TestTen99PolicyResponseBase(unittest.TestCase):
    def setUp(self):
        self.code = 200
        self.headers = {
            "Ten99Policy-Idempotent-Key": "idempotency_key_value",
            "request-id": "request_id_value",
        }
        self.response_base = Ten99PolicyResponseBase(self.code, self.headers)

    def test_init(self):
        self.assertEqual(self.response_base.code, self.code)
        self.assertEqual(self.response_base.headers, self.headers)

    def test_idempotency_key(self):
        self.assertEqual(self.response_base.idempotency_key, "idempotency_key_value")

    def test_idempotency_key_missing(self):
        response_base = Ten99PolicyResponseBase(self.code, {})
        self.assertIsNone(response_base.idempotency_key)

    def test_request_id(self):
        self.assertEqual(self.response_base.request_id, "request_id_value")

    def test_request_id_missing(self):
        response_base = Ten99PolicyResponseBase(self.code, {})
        self.assertIsNone(response_base.request_id)


class TestTen99PolicyResponse(unittest.TestCase):
    def setUp(self):
        self.body = json.dumps({"key": "value"})
        self.code = 200
        self.headers = {"Content-Type": "application/json"}
        self.response = Ten99PolicyResponse(self.body, self.code, self.headers)

    def test_init(self):
        self.assertEqual(self.response.body, self.body)
        self.assertEqual(self.response.code, self.code)
        self.assertEqual(self.response.headers, self.headers)
        self.assertEqual(self.response.data, OrderedDict([("key", "value")]))

    def test_data_property(self):
        self.assertEqual(self.response.data, OrderedDict([("key", "value")]))


class TestTen99PolicyStreamResponse(unittest.TestCase):
    def setUp(self):
        self.io = "streaming data"
        self.code = 200
        self.headers = {"Content-Type": "application/octet-stream"}
        self.response = Ten99PolicyStreamResponse(self.io, self.code, self.headers)

    def test_init(self):
        self.assertEqual(self.response.io, self.io)
        self.assertEqual(self.response.code, self.code)
        self.assertEqual(self.response.headers, self.headers)


if __name__ == "__main__":
    unittest.main()



import unittest
import textwrap
from unittest import mock
from unittest.mock import patch, MagicMock, PropertyMock
from unittest.mock import Mock
import urllib
from io import BytesIO

from ten99policy.http_client import (
    HTTPClient,
    RequestsClient,
    UrlFetchClient,
    PycurlClient,
    Urllib2Client,
    new_default_http_client,
)
from ten99policy import error

# Mock dependencies that may not be available during testing
try:
    import requests
except ImportError:
    requests = None

try:
    import pycurl
except ImportError:
    pycurl = None

try:
    from google.appengine.api import urlfetch
except ImportError:
    urlfetch = None


class TestHTTPClientBase(unittest.TestCase):
    def setUp(self):
        self.client = HTTPClient()

    def test_initialization_without_proxy(self):
        client = HTTPClient()
        self.assertTrue(client._verify_ssl_certs)
        self.assertIsNone(client._proxy)

    def test_initialization_with_proxy_string(self):
        proxy = "http://proxy.example.com:8080"
        client = HTTPClient(proxy=proxy)
        expected_proxy = {"http": proxy, "https": proxy}
        self.assertEqual(client._proxy, expected_proxy)

    def test_initialization_with_proxy_dict(self):
        proxy = {
            "http": "http://proxy.example.com:8080",
            "https": "https://proxy.example.com:8443",
        }
        client = HTTPClient(proxy=proxy)
        self.assertEqual(client._proxy, proxy)

    def test_initialization_with_invalid_proxy_type(self):
        with self.assertRaises(ValueError):
            HTTPClient(proxy=123)  # Invalid proxy type

    def test_request_with_retries_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.client.request_with_retries("GET", "http://example.com", {})

    def test_request_stream_with_retries_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.client.request_stream_with_retries("GET", "http://example.com", {})

    def test_should_retry_max_retries_exceeded(self):
        response = (None, 500, {})
        api_connection_error = mock.Mock()
        api_connection_error.should_retry = True
        result = self.client._should_retry(response, api_connection_error, 10)
        self.assertFalse(result)

    def test_should_retry_response_header_force_retry_false(self):
        response = (None, 200, {"ten99policy-should-retry": "false"})
        api_connection_error = None
        result = self.client._should_retry(response, api_connection_error, 1)
        self.assertFalse(result)

    @patch("ten99policy.http_client.time.sleep", return_value=None)
    def test_request_with_retries_success(self, mock_sleep):
        self.client._request_with_retries_internal = mock.Mock(return_value="success")
        result = self.client.request_with_retries("GET", "http://example.com", {})
        self.assertEqual(result, "success")
        self.client._request_with_retries_internal.assert_called_once_with(
            "GET", "http://example.com", {}, None, is_streaming=False
        )

    @patch("ten99policy.http_client.time.sleep", return_value=None)
    def test_request_with_retries_retry_then_success(self, mock_sleep):
        self.client._should_retry = mock.Mock(side_effect=[True, False])
        self.client.request = mock.Mock(
            side_effect=[
                (None, 500, {}),  # First response triggers a retry
                (b'{"success": true}', 200, {}),  # Second response is successful
            ]
        )
        result = self.client.request_with_retries("GET", "http://example.com", {})
        self.assertEqual(result, (b'{"success": true}', 200, {}))
        self.assertEqual(self.client.request.call_count, 2)

    def test_sleep_time_seconds_exponential_backoff(self):
        with patch(
            "ten99policy.http_client.HTTPClient._retry_after_header", return_value=None
        ):
            sleep_time = self.client._sleep_time_seconds(1)
            expected = max(0.5, 0.5 * (1 + 0.5))  # 0.5 * (1 + 0.5) = 0.75
            self.assertGreaterEqual(sleep_time, 0.5)
            self.assertLessEqual(sleep_time, 2)

    def test_sleep_time_seconds_with_retry_after(self):
        with patch(
            "ten99policy.http_client.HTTPClient._retry_after_header", return_value=3
        ):
            sleep_time = self.client._sleep_time_seconds(1)
            self.assertGreaterEqual(sleep_time, 3)

    def test_add_jitter_time(self):
        with patch("ten99policy.http_client.random.uniform", return_value=0.5):
            sleep_time = self.client._add_jitter_time(2)
            self.assertEqual(sleep_time, 2 * 0.5 * (1 + 0.5))  # 2 * 0.5 * 1.5 = 1.5

    def test_close_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.client.close()


class TestHTTPClientShouldRetry(unittest.TestCase):
    def setUp(self):
        self.client = HTTPClient()
        self.client._max_network_retries = Mock(return_value=3)

    def test_max_retries_exceeded(self):
        result = self.client._should_retry(None, None, 3)
        self.assertFalse(result)

    def test_response_none_should_retry(self):
        api_connection_error = Mock(should_retry=True)
        result = self.client._should_retry(None, api_connection_error, 1)
        self.assertTrue(result)

    def test_response_none_should_not_retry(self):
        api_connection_error = Mock(should_retry=False)
        result = self.client._should_retry(None, api_connection_error, 1)
        self.assertFalse(result)

    def test_should_retry_header_true(self):
        response = (None, 200, {"ten99policy-should-retry": "true"})
        result = self.client._should_retry(response, None, 1)
        self.assertTrue(result)

    def test_should_retry_header_false(self):
        response = (None, 200, {"ten99policy-should-retry": "false"})
        result = self.client._should_retry(response, None, 1)
        self.assertFalse(result)

    def test_conflict_status_code(self):
        response = (None, 409, {})
        result = self.client._should_retry(response, None, 1)
        self.assertTrue(result)

    def test_internal_server_error(self):
        response = (None, 500, {})
        result = self.client._should_retry(response, None, 1)
        self.assertTrue(result)

    def test_other_5xx_error(self):
        response = (None, 503, {})
        result = self.client._should_retry(response, None, 1)
        self.assertTrue(result)

    def test_non_retryable_status_code(self):
        response = (None, 400, {})
        result = self.client._should_retry(response, None, 1)
        self.assertFalse(result)

    def test_should_retry_header_overrides_status_code(self):
        response = (None, 500, {"ten99policy-should-retry": "false"})
        result = self.client._should_retry(response, None, 1)
        self.assertFalse(result)

    @unittest.skip("Temporarily skipping this test")
    def test_case_insensitive_headers(self):
        response = (None, 200, {"TEN99POLICY-SHOULD-RETRY": "true"})
        result = self.client._should_retry(response, None, 1)
        self.assertTrue(result)


class TestRequestsClient(unittest.TestCase):
    def setUp(self):
        self.mock_session = mock.Mock()
        self.client = RequestsClient(timeout=30, session=self.mock_session)

    def test_initialization(self):
        self.assertEqual(self.client._timeout, 30)
        self.assertEqual(self.client._session, self.mock_session)
        self.assertTrue(self.client._verify_ssl_certs)
        self.assertIsNone(self.client._proxy)

    @patch("ten99policy.http_client.requests.Session")
    def test_initialization_with_default_session(self, mock_session):
        client = RequestsClient()
        mock_session.assert_not_called()
        self.assertIs(client._session, client._session)

    def test_request_internal_get(self):
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.content = b'{"success": true}'
        mock_response.headers = {}
        self.client._thread_local.session = self.mock_session
        self.mock_session.request.return_value = mock_response

        content, status_code, headers = self.client._request_internal(
            "get", "http://example.com", {}, None, False
        )

        self.mock_session.request.assert_called_with(
            "get", "http://example.com", headers={}, data=None, timeout=30, verify=False
        )
        self.assertEqual(content, mock_response.content)
        self.assertEqual(status_code, 200)
        self.assertEqual(headers, mock_response.headers)

    def test_request_internal_post(self):
        mock_response = mock.Mock()
        mock_response.status_code = 201
        mock_response.content = b'{"created": true}'
        mock_response.headers = {}
        self.client._thread_local.session = self.mock_session
        self.mock_session.request.return_value = mock_response

        content, status_code, headers = self.client._request_internal(
            "post", "http://example.com", {}, "data", False
        )

        self.mock_session.request.assert_called_with(
            "post",
            "http://example.com",
            headers={},
            data="data",
            timeout=30,
            verify=False,
        )
        self.assertEqual(content, mock_response.content)
        self.assertEqual(status_code, 201)
        self.assertEqual(headers, mock_response.headers)

    def test_request_internal_streaming(self):
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.raw = "streaming content"
        mock_response.headers = {}
        self.client._thread_local.session = self.mock_session
        self.mock_session.request.return_value = mock_response

        content, status_code, headers = self.client._request_internal(
            "get", "http://example.com", {}, None, True
        )

        self.mock_session.request.assert_called_with(
            "get",
            "http://example.com",
            headers={},
            data=None,
            timeout=30,
            verify=False,
            stream=True,
        )
        self.assertEqual(content, mock_response.raw)
        self.assertEqual(status_code, 200)
        self.assertEqual(headers, mock_response.headers)

    def test_handle_request_error_ssl_error(self):
        ssl_error = requests.exceptions.SSLError("SSL certificate error")
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._handle_request_error(ssl_error)
        self.assertIn(
            "Could not verify Ten99Policy's SSL certificate", str(context.exception)
        )
        self.assertFalse(context.exception.should_retry)

    def test_handle_request_error_timeout(self):
        timeout_error = requests.exceptions.Timeout("Request timed out")
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._handle_request_error(timeout_error)
        self.assertIn(
            "Unexpected error communicating with ten99policy", str(context.exception)
        )
        self.assertTrue(context.exception.should_retry)

    def test_handle_request_error_connection_error(self):
        connection_error = requests.exceptions.ConnectionError("Connection failed")
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._handle_request_error(connection_error)
        self.assertIn(
            "Unexpected error communicating with ten99policy", str(context.exception)
        )
        self.assertTrue(context.exception.should_retry)

    def test_handle_request_error_generic_request_exception(self):
        request_exception = requests.exceptions.RequestException("Generic error")
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._handle_request_error(request_exception)
        self.assertIn(
            "Unexpected error communicating with ten99policy", str(context.exception)
        )
        self.assertFalse(context.exception.should_retry)

    def test_close_session(self):
        self.client._thread_local.session = self.mock_session
        self.client.close()
        self.mock_session.close.assert_called_once()


@unittest.skipIf(urlfetch is None, "urlfetch not available")
class TestUrlFetchClient(unittest.TestCase):
    def setUp(self):
        self.client = UrlFetchClient()

    def test_initialization_with_proxy_raises_error(self):
        with self.assertRaises(ValueError):
            UrlFetchClient(proxy="http://proxy.example.com")

    @patch("ten99policy.http_client.urlfetch.fetch")
    def test_request_internal_get(self, mock_fetch):
        mock_result = mock.Mock()
        mock_result.content = b'{"success": true}'
        mock_result.status_code = 200
        mock_result.headers = {}
        mock_fetch.return_value = mock_result

        content, status_code, headers = self.client._request_internal(
            "get", "http://example.com", {}, None, False
        )

        mock_fetch.assert_called_with(
            url="http://example.com",
            method="get",
            headers={},
            validate_certificate=True,
            deadline=55,
            payload=None,
        )
        self.assertEqual(content, mock_result.content)
        self.assertEqual(status_code, 200)
        self.assertEqual(headers, mock_result.headers)

    @patch("ten99policy.http_client.urlfetch.fetch")
    def test_request_internal_post_streaming(self, mock_fetch):
        mock_result = mock.Mock()
        mock_result.content = b"streaming content"
        mock_result.status_code = 200
        mock_result.headers = {}
        mock_fetch.return_value = mock_result

        content, status_code, headers = self.client._request_internal(
            "post", "http://example.com", {}, "data", True
        )

        mock_fetch.assert_called_with(
            url="http://example.com",
            method="post",
            headers={},
            validate_certificate=True,
            deadline=55,
            payload="data",
        )
        self.assertEqual(content, b"streaming content")
        self.assertEqual(status_code, 200)
        self.assertEqual(headers, mock_result.headers)

    @patch("ten99policy.http_client.urlfetch.fetch")
    def test_handle_request_error_invalid_url(self, mock_fetch):
        mock_fetch.side_effect = urlfetch.InvalidURLError()
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._request_internal("get", "invalid:url", {}, None, False)
        self.assertIn("attempted to fetch an invalid URL", str(context.exception))

    @patch("ten99policy.http_client.urlfetch.fetch")
    def test_handle_request_error_download_error(self, mock_fetch):
        mock_fetch.side_effect = urlfetch.DownloadError()
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._request_internal("get", "http://example.com", {}, None, False)
        self.assertIn(
            "problem retrieving data from ten99policy", str(context.exception)
        )

    def test_close_no_op(self):
        # UrlFetchClient.close is a no-op
        try:
            self.client.close()
        except Exception as e:
            self.fail(f"UrlFetchClient.close() raised an exception {e}")


@unittest.skipIf(pycurl is None, "pycurl not available")
class TestPycurlClient(unittest.TestCase):
    def setUp(self):
        self.patcher = patch("ten99policy.http_client.pycurl.Curl")
        self.mock_curl_class = self.patcher.start()
        self.mock_curl = mock.Mock()
        self.mock_curl_class.return_value = self.mock_curl
        self.client = PycurlClient()

    def tearDown(self):
        self.patcher.stop()

    def test_initialization_with_proxy(self):
        proxy = {"http": "http://proxy.example.com:8080"}
        client = PycurlClient(proxy=proxy)
        parsed_proxy = mock.Mock()
        with patch("ten99policy.http_client.urlparse", return_value=parsed_proxy):
            client._get_proxy("http://example.com")
            self.mock_curl.setopt.assert_called_with(
                pycurl.PROXY, parsed_proxy.hostname
            )

    @patch("ten99policy.http_client.util.io.BytesIO")
    def test_request_internal_get(self, mock_bytesio):
        mock_response = mock.Mock()
        mock_response.getvalue.return_value = b'{"success": true}'
        mock_write = mock_bytesio.return_value.write
        mock_header_write = mock_bytesio.return_value.write
        self.mock_curl.perform.return_value = None
        self.mock_curl.getinfo.return_value = 200
        mock_curl_instance = self.mock_curl

        self.mock_curl.getinfo.return_value = 200
        mock_bytesio.side_effect = [mock.Mock(), mock.Mock()]
        content, status_code, headers = self.client._request_internal(
            "get", "http://example.com", {}, None, False
        )

        self.mock_curl.setopt.assert_any_call(pycurl.HTTPGET, 1)
        self.mock_curl.setopt.assert_any_call(pycurl.URL, "http://example.com")
        self.assertTrue(mock_curl_instance.perform.called)
        self.assertEqual(status_code, 200)

    @patch("ten99policy.http_client.util.io.BytesIO")
    def test_request_internal_post_streaming(self, mock_bytesio):
        mock_response = mock.Mock()
        mock_response.getvalue.return_value = b"streaming content"
        self.mock_curl.perform.return_value = None
        self.mock_curl.getinfo.return_value = 200
        mock_bytesio.side_effect = [mock.Mock(), mock.Mock()]
        content, status_code, headers = self.client._request_internal(
            "post", "http://example.com", {}, "data", True
        )

        self.mock_curl.setopt.assert_any_call(pycurl.POST, 1)
        self.mock_curl.setopt.assert_any_call(pycurl.POSTFIELDS, "data")
        self.mock_curl.setopt.assert_any_call(pycurl.URL, "http://example.com")
        self.assertTrue(self.mock_curl.perform.called)
        self.assertEqual(status_code, 200)

    def test_handle_request_error_connection_error(self):
        self.mock_curl.perform.side_effect = pycurl.error(
            pycurl.E_COULDNT_CONNECT, "Connection failed"
        )
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._request_internal("get", "http://example.com", {}, None, False)
        self.assertIn("Could not connect to ten99policy", str(context.exception))
        self.assertTrue(context.exception.should_retry)

    def test_handle_request_error_ssl_error(self):
        self.mock_curl.perform.side_effect = pycurl.error(
            pycurl.E_SSL_CACERT, "SSL cert error"
        )
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._request_internal("get", "http://example.com", {}, None, False)
        self.assertIn(
            "Could not verify Ten99Policy's SSL certificate", str(context.exception)
        )
        self.assertFalse(context.exception.should_retry)

    @patch("ten99policy.http_client.util.io.BytesIO")
    def test_request_internal_custom_method(self, mock_bytesio):
        # Example of a PUT request with custom method
        mock_bytesio.side_effect = [mock.Mock(), mock.Mock()]
        self.mock_curl.getinfo.return_value = 204
        self.client._request_internal(
            "put", "http://example.com/resource", {}, None, False
        )
        self.mock_curl.setopt.assert_any_call(pycurl.CUSTOMREQUEST, "PUT")
        self.assertTrue(self.mock_curl.perform.called)


@unittest.skipIf(requests is None, "requests not available")
class TestRequestsClientWithProxy(unittest.TestCase):
    def setUp(self):
        self.mock_session = mock.Mock()
        self.proxy = {
            "http": "http://proxy.example.com:8080",
            "https": "https://proxy.example.com:8443",
        }
        self.client = RequestsClient(proxy=self.proxy)

    @patch("ten99policy.http_client.requests.Session")
    def test_initialization_with_proxy(self, mock_session):
        client = RequestsClient(proxy=self.proxy)
        self.assertEqual(client._proxy, self.proxy)

    @patch("ten99policy.http_client.requests.Session")
    def test_request_internal_with_proxy(self, mock_session):
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.content = b'{"success": true}'
        mock_response.headers = {}
        self.client._thread_local.session = self.mock_session
        self.mock_session.request.return_value = mock_response

        content, status_code, headers = self.client._request_internal(
            "get", "http://example.com", {}, None, False
        )

        expected_proxies = self.proxy
        self.mock_session.request.assert_called_with(
            "get",
            "http://example.com",
            headers={},
            data=None,
            timeout=80,
            verify=False,
            proxies=expected_proxies,
        )
        self.assertEqual(content, mock_response.content)
        self.assertEqual(status_code, 200)
        self.assertEqual(headers, mock_response.headers)


class TestUrllib2Client(unittest.TestCase):
    def setUp(self):
        self.patcher = patch("ten99policy.http_client.urllib.request.OpenerDirector")
        self.mock_opener = self.patcher.start()
        self.client = Urllib2Client()

    def tearDown(self):
        self.patcher.stop()

    def test_initialization_without_proxy(self):
        client = Urllib2Client()
        self.assertIsNone(client._opener)

    def test_initialization_with_proxy(self):
        proxy = {
            "http": "http://proxy.example.com:8080",
            "https": "https://proxy.example.com:8443",
        }
        with patch(
            "ten99policy.http_client.urllib.request.ProxyHandler"
        ) as mock_proxy_handler, patch(
            "ten99policy.http_client.urllib.request.build_opener"
        ) as mock_build_opener:
            mock_proxy_handler.return_value = MagicMock()
            mock_build_opener.return_value = MagicMock()
            client = Urllib2Client(proxy=proxy)
            mock_proxy_handler.assert_called_with(proxy)
            mock_build_opener.assert_called_with(mock_proxy_handler.return_value)
            self.assertEqual(client._opener, mock_build_opener.return_value)

    @patch("ten99policy.http_client.urllib.request.Request")
    @patch("ten99policy.http_client.urllib.request.urlopen")
    def test_request_internal_get(self, mock_urlopen, mock_request):
        mock_response = mock.Mock()
        mock_response.read.return_value = b'{"success": true}'
        mock_response.code = 200
        mock_response.info.return_value = {"Content-Type": "application/json"}
        mock_urlopen.return_value = mock_response

        content, status_code, headers = self.client._request_internal(
            "get", "http://example.com", {}, None, False
        )

        mock_request.assert_called_with("http://example.com", None, {})
        mock_urlopen.assert_called_with(mock_request.return_value)
        self.assertEqual(content, b'{"success": true}')
        self.assertEqual(status_code, 200)

        # Convert both dictionaries to lowercase keys for comparison
        lowercase_headers = {k.lower(): v for k, v in headers.items()}
        expected_headers = {"content-type": "application/json"}
        self.assertEqual(lowercase_headers, expected_headers)

    @patch(
        "ten99policy.http_client.urllib.request.urlopen",
        side_effect=urllib.error.URLError("Timeout"),
    )
    def test_handle_request_error_urllib_error(self, mock_urlopen):
        with self.assertRaises(error.APIConnectionError) as context:
            self.client._request_internal("get", "http://example.com", {}, None, False)
        self.assertIn(
            "Unexpected error communicating with ten99policy", str(context.exception)
        )

    @patch("ten99policy.http_client.urllib.request.Request")
    @patch("ten99policy.http_client.urllib.request.urlopen")
    def test_request_internal_post_streaming(self, mock_urlopen, mock_request):
        mock_response = mock.Mock()
        mock_response.read.return_value = b'{"created": true}'
        mock_response.code = 201
        mock_response.info.return_value = {"Content-Type": "application/json"}
        mock_urlopen.return_value = mock_response

        content, status_code, headers = self.client._request_internal(
            "post", "http://example.com/resource", {}, "data", True
        )

        mock_request.assert_called_with("http://example.com/resource", b"data", {})
        mock_urlopen.assert_called_with(mock_request.return_value)
        self.assertEqual(content, mock_urlopen.return_value)
        self.assertEqual(status_code, 201)
        self.assertEqual(headers, {"content-type": "application/json"})
        self.assertEqual(content.read(), b'{"created": true}')

    def test_close_no_op(self):
        # Urllib2Client.close is a no-op
        try:
            self.client.close()
        except Exception as e:
            self.fail(f"Urllib2Client.close() raised an exception {e}")


# Create a mock urlfetch module with the required error classes
mock_urlfetch = Mock()
mock_urlfetch.InvalidURLError = type("InvalidURLError", (Exception,), {})
mock_urlfetch.DownloadError = type("DownloadError", (Exception,), {})
mock_urlfetch.ResponseTooLargeError = type("ResponseTooLargeError", (Exception,), {})


class TestUrlFetchClientHandleRequestError(unittest.TestCase):
    @patch("ten99policy.http_client.urlfetch", mock_urlfetch)
    def setUp(self):
        self.client = UrlFetchClient()
        self.url = "https://api.ten99policy.com/v1/example"

    @patch("ten99policy.http_client.urlfetch", mock_urlfetch)
    def test_invalid_url_error(self):
        e = mock_urlfetch.InvalidURLError()
        expected_msg = (
            "The Ten99Policy library attempted to fetch an "
            "invalid URL ('https://api.ten99policy.com/v1/example'). This is likely due to a bug "
            "in the Ten99Policy Python bindings. Please let us know "
            "at support@1099policy.com."
        )
        expected_msg = textwrap.fill(expected_msg) + f"\n\n(Network error: {str(e)})"

        with self.assertRaises(error.APIConnectionError) as cm:
            self.client._handle_request_error(e, self.url)

        self.assertEqual(str(cm.exception), expected_msg)

    @patch("ten99policy.http_client.urlfetch", mock_urlfetch)
    def test_download_error(self):
        e = mock_urlfetch.DownloadError()
        expected_msg = "There was a problem retrieving data from ten99policy."
        expected_msg = textwrap.fill(expected_msg) + f"\n\n(Network error: {str(e)})"

        with self.assertRaises(error.APIConnectionError) as cm:
            self.client._handle_request_error(e, self.url)

        self.assertEqual(str(cm.exception), expected_msg)

    @patch("ten99policy.http_client.urlfetch", mock_urlfetch)
    def test_response_too_large_error(self):
        e = mock_urlfetch.ResponseTooLargeError()
        expected_msg = (
            "There was a problem receiving all of your data from "
            "ten99policy.  This is likely due to a bug in ten99policy. "
            "Please let us know at support@1099policy.com."
        )
        expected_msg = textwrap.fill(expected_msg) + f"\n\n(Network error: {str(e)})"

        with self.assertRaises(error.APIConnectionError) as cm:
            self.client._handle_request_error(e, self.url)

        self.assertEqual(str(cm.exception), expected_msg)

    @patch("ten99policy.http_client.urlfetch", mock_urlfetch)
    def test_unexpected_error(self):
        e = Exception("Unexpected error")
        expected_msg = (
            "Unexpected error communicating with ten99policy. If this "
            "problem persists, let us know at support@1099policy.com."
        )
        expected_msg = textwrap.fill(expected_msg) + f"\n\n(Network error: {str(e)})"

        with self.assertRaises(error.APIConnectionError) as cm:
            self.client._handle_request_error(e, self.url)

        self.assertEqual(str(cm.exception), expected_msg)


class TestNewDefaultHttpClient(unittest.TestCase):
    @patch("ten99policy.http_client.urlfetch", MagicMock())
    def test_new_default_http_client_urlfetch_available(self):
        client = new_default_http_client()
        self.assertIsInstance(client, UrlFetchClient)

    @patch("ten99policy.http_client.urlfetch", None)
    @patch("ten99policy.http_client.requests", MagicMock())
    def test_new_default_http_client_requests_available(self):
        client = new_default_http_client()
        self.assertIsInstance(client, RequestsClient)

    @patch("ten99policy.http_client.urlfetch", None)
    @patch("ten99policy.http_client.requests", None)
    @patch("ten99policy.http_client.pycurl", MagicMock())
    def test_new_default_http_client_pycurl_available(self):
        client = new_default_http_client()
        self.assertIsInstance(client, PycurlClient)

    @patch("ten99policy.http_client.urlfetch", None)
    @patch("ten99policy.http_client.requests", None)
    @patch("ten99policy.http_client.pycurl", None)
    @patch("ten99policy.http_client.urllib.request", MagicMock())
    def test_new_default_http_client_urllib2_fallback(self):
        with patch("ten99policy.http_client.warnings") as mock_warnings:
            client = new_default_http_client()
            self.assertIsInstance(client, Urllib2Client)
            mock_warnings.warn.assert_called_once()

    @patch("ten99policy.http_client.urlfetch", None)
    @patch("ten99policy.http_client.requests", MagicMock())
    def test_new_default_http_client_requests_new_enough(self):
        with patch("ten99policy.http_client.requests") as mock_requests:
            type(mock_requests).version = PropertyMock(return_value="0.8.8")
            client = new_default_http_client()
            self.assertIsInstance(client, RequestsClient)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function

# flake8: noqa

from ten99policy.api_resources.abstract.api_resource import APIResource
from ten99policy.api_resources.abstract.singleton_api_resource import(SingletonAPIResource)
from ten99policy.api_resources.abstract.createable_api_resource import(CreateableAPIResource)
from ten99policy.api_resources.abstract.updateable_api_resource import(UpdateableAPIResource)
from ten99policy.api_resources.abstract.deletable_api_resource import(DeletableAPIResource)
from ten99policy.api_resources.abstract.listable_api_resource import(ListableAPIResource)
from ten99policy.api_resources.abstract.custom_method import custom_method
from ten99policy.api_resources.abstract.nested_resource_class_methods import(nested_resource_class_methods)



from ten99policy.api_resources.abstract import CreateableAPIResource
from ten99policy.api_resources.abstract import ListableAPIResource
from ten99policy.api_resources.abstract import UpdateableAPIResource


class Quotes(
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "quotes"



# tests/test_multipart_data_generator.py
import unittest
import io
from ten99policy.multipart_data_generator import MultipartDataGenerator


class TestMultipartDataGenerator(unittest.TestCase):

    def setUp(self):
        self.generator = MultipartDataGenerator()

    def test_add_params_with_string(self):
        # Test adding params with string values
        params = {"key1": "value1", "key2": "value2"}
        self.generator.add_params(params)
        post_data = self.generator.get_post_data().decode("utf-8")

        # Check if the params are correctly added
        self.assertIn('Content-Disposition: form-data; name="key1"', post_data)
        self.assertIn('Content-Disposition: form-data; name="key2"', post_data)
        self.assertIn("value1", post_data)
        self.assertIn("value2", post_data)

    def test_add_params_with_none_value(self):
        # Test adding params with None value
        params = {"key1": "value1", "key2": None}
        self.generator.add_params(params)
        post_data = self.generator.get_post_data().decode("utf-8")

        # Check if the None value is skipped
        self.assertIn('Content-Disposition: form-data; name="key1"', post_data)
        self.assertIn("value1", post_data)
        self.assertNotIn('Content-Disposition: form-data; name="key2"', post_data)

    def test_add_params_with_int_value(self):
        # Test adding params with integer value
        params = {"key1": 123}
        self.generator.add_params(params)
        post_data = self.generator.get_post_data().decode("utf-8")

        # Check if the integer value is correctly added
        self.assertIn('Content-Disposition: form-data; name="key1"', post_data)
        self.assertIn("123", post_data)

    def test_add_params_with_unicode(self):
        # Test adding params with unicode value
        params = {"key1": "值"}
        self.generator.add_params(params)
        post_data = self.generator.get_post_data().decode("utf-8")

        # Check if the unicode value is correctly added
        self.assertIn('Content-Disposition: form-data; name="key1"', post_data)
        self.assertIn("值", post_data)

    def test_add_params_with_file(self):
        # Create a mock file-like object
        mock_file = io.BytesIO(b"file content")
        mock_file.name = "test_file.txt"
        params = {"file": mock_file}

        self.generator.add_params(params)
        post_data = self.generator.get_post_data().decode("utf-8")

        # Check if the file is correctly added
        self.assertIn(
            'Content-Disposition: form-data; name="file"; filename="test_file.txt"',
            post_data,
        )
        self.assertIn("Content-Type: application/octet-stream", post_data)
        self.assertIn("file content", post_data)

    def test_get_post_data(self):
        params = {"key1": "value1"}
        self.generator.add_params(params)
        post_data = self.generator.get_post_data()

        # Check if the post data is not empty
        self.assertGreater(len(post_data), 0)

    def test_boundary_initialization(self):
        # Check if the boundary is initialized correctly
        boundary = self.generator.boundary
        self.assertIsInstance(boundary, int)

    def test_write_method_with_string(self):
        self.generator._write("test")
        post_data = self.generator.get_post_data().decode("utf-8")
        self.assertIn("test", post_data)

    def test_write_method_with_bytes(self):
        self.generator._write(b"test bytes")
        post_data = self.generator.get_post_data().decode("utf-8")
        self.assertIn("test bytes", post_data)

    def test_write_file_method(self):
        # Create a mock file-like object
        mock_file = io.BytesIO(b"file content")
        mock_file.name = "test_file.txt"
        self.generator._write_file(mock_file)

        post_data = self.generator.get_post_data().decode("utf-8")
        self.assertIn("file content", post_data)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function

import ten99policy
from ten99policy.six import python_2_unicode_compatible


@python_2_unicode_compatible
class Ten99PolicyError(Exception):
    def __init__(
        self,
        message=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
    ):
        super(Ten99PolicyError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as utf-8. "
                    "Please report to support@1099policy.com>"
                )

        self._message = message

        if isinstance(message, dict):
            try:
                error_message = ", ".join(
                    f"{key}: {value}" for key, value in message.items()
                )
            except Exception:
                error_message = "Bad request"

            self._message = error_message

        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}
        self.code = code
        self.request_id = self.headers.get("request-id", None)
        self.error = self.construct_error_object()

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.request_id is not None:
            return "Request {0}: {1}".format(self.request_id, msg)
        else:
            return msg

    # Returns the underlying `Exception` (base class) message, which is usually
    # the raw message returned by Ten99Policy's API. This was previously available
    # in python2 via `error.message`. Unlike `str(error)`, it omits "Request
    # req_..." from the beginning of the string.
    @property
    def message(self):
        return self._message

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
            self.request_id,
        )

    def construct_error_object(self):
        if (
            self.json_body is None
            or "error" not in self.json_body
            or not isinstance(self.json_body["error"], dict)
        ):
            return None

        return ten99policy.api_resources.error_object.ErrorObject.construct_from(
            self.json_body["error"], ten99policy.api_key
        )


class APIError(Ten99PolicyError):
    pass


class APIConnectionError(Ten99PolicyError):
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.should_retry = should_retry


class Ten99PolicyErrorWithParamCode(Ten99PolicyError):
    def __repr__(self):
        return "%s(message=%r, param=%r, code=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.param,
            self.code,
            self.http_status,
            self.request_id,
        )


class IdempotencyError(Ten99PolicyError):
    pass


class InvalidRequestError(Ten99PolicyErrorWithParamCode):
    def __init__(
        self,
        message,
        param,
        code=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(InvalidRequestError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.param = param


class AuthenticationError(Ten99PolicyError):
    pass


class PermissionError(Ten99PolicyError):
    pass


class RateLimitError(Ten99PolicyError):
    pass


class GeneralError(Ten99PolicyError):
    pass


class InvalidApiKeyError(Ten99PolicyError):
    pass


class AuthTokenExpiredError(Ten99PolicyError):
    pass


class ResourceNotFoundError(Ten99PolicyError):
    pass


class DatabaseOperationalError(Ten99PolicyError):
    pass


class InsufficientPermissionsError(Ten99PolicyError):
    pass


class NoTenantFoundError(Ten99PolicyError):
    pass


class InvalidWebhookSignatureError(Ten99PolicyError):
    pass


class InvalidInputError(Ten99PolicyError):
    pass


class BadRequestError(Ten99PolicyError):
    pass


class InvalidQuoteIdError(Ten99PolicyError):
    pass


class InvalidSessionIdError(Ten99PolicyError):
    pass


class SessionExpiredError(Ten99PolicyError):
    pass


class ApplicationAlreadyCompleteError(Ten99PolicyError):
    pass


class EffectiveDateInvalidError(Ten99PolicyError):
    pass


class EndDateInvalidError(Ten99PolicyError):
    pass


class InvalidContractorIdError(Ten99PolicyError):
    pass


class InvalidPolicyIdError(Ten99PolicyError):
    pass


class InvalidJobIdError(Ten99PolicyError):
    pass


class JobIsUsedError(Ten99PolicyError):
    pass


class InvalidAssignmentIdError(Ten99PolicyError):
    pass


class InvoiceAlreadyPaidError(Ten99PolicyError):
    pass


class MissingCertificatesError(Ten99PolicyError):
    pass


class InvalidFileTypeError(Ten99PolicyError):
    pass


class DuplicateEmailError(Ten99PolicyError):
    pass


class InvalidEntityIdError(Ten99PolicyError):
    pass


class InvalidEventIdError(Ten99PolicyError):
    pass


class InvalidPaycycleStartdateError(Ten99PolicyError):
    pass


class InvalidPaycycleEnddateError(Ten99PolicyError):
    pass


class NoActivePolicyError(Ten99PolicyError):
    pass


class AgencyPayInvoiceExistsError(Ten99PolicyError):
    pass


class InvalidInvoiceIdError(Ten99PolicyError):
    pass


class InvoiceUneditableAlreadyExistsError(Ten99PolicyError):
    pass


class MissingJobCategoryCodeError(Ten99PolicyError):
    pass


class InvalidJobCategoryCodeError(Ten99PolicyError):
    pass


class JobCategoryNotApprovedError(Ten99PolicyError):
    pass


class InvoiceExistsError(Ten99PolicyError):
    pass


class CustomApplicationsDisabledError(Ten99PolicyError):
    pass


class ContractorHasMatchingPolicyError(Ten99PolicyError):
    pass


class CantUpdateCreateNewQuoteError(Ten99PolicyError):
    pass


class PolicyAlreadyExistsError(Ten99PolicyError):
    pass


class InvalidWebhookEndpointIdError(Ten99PolicyError):
    pass


class MissingContractorIdError(Ten99PolicyError):
    pass


class MissingCertificateError(Ten99PolicyError):
    pass


class NoInsuranceRequirementFoundError(Ten99PolicyError):
    pass


class InvalidMimeTypeError(Ten99PolicyError):
    pass


class InvalidFileSizeError(Ten99PolicyError):
    pass


class ContractorBlockedForWritesError(Ten99PolicyError):
    pass



import unittest
from unittest.mock import patch, Mock
import json
import datetime
from collections import OrderedDict

from ten99policy.api_requestor import (
    APIRequestor,
    _encode_datetime,
    _encode_nested_dict,
    _api_encode,
    _build_api_url,
)
from ten99policy import error
from ten99policy.http_client import RequestsClient


class TestAPIRequestor(unittest.TestCase):

    def setUp(self):
        self.api_key = "test_key"
        self.api_base = "https://api.test.com"
        self.requestor = APIRequestor(key=self.api_key, api_base=self.api_base)

    def test_init(self):
        self.assertEqual(self.requestor.api_key, self.api_key)
        self.assertEqual(self.requestor.api_base, self.api_base)

    @patch("ten99policy.api_requestor.http_client.RequestsClient")
    def test_init_default_http_client(self, mock_client):
        requestor = APIRequestor()
        self.assertIsInstance(requestor._client, RequestsClient)

    def test_encode_datetime(self):
        dt = datetime.datetime(2023, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)
        encoded = _encode_datetime(dt)
        self.assertEqual(encoded, 1672574400)

    def test_encode_nested_dict(self):
        data = {"a": 1, "b": 2}
        encoded = _encode_nested_dict("test", data)
        self.assertEqual(encoded, OrderedDict([("test[a]", 1), ("test[b]", 2)]))

    def test_api_encode(self):
        data = {
            "string": "value",
            "int": 42,
            "list": [1, 2, 3],
            "dict": {"a": 1, "b": 2},
            "datetime": datetime.datetime(
                2023, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc
            ),
        }
        encoded = list(_api_encode(data))
        self.assertIn(("string", "value"), encoded)
        self.assertIn(
            ("int", 42), encoded
        )  # Note: We're now expecting an integer, not a string
        self.assertIn(("list[0]", 1), encoded)
        self.assertIn(("dict[a]", 1), encoded)
        self.assertIn(("datetime", 1672574400), encoded)

    def test_build_api_url(self):
        url = "https://api.test.com/v1/endpoint"
        query = "param1=value1&param2=value2"
        built_url = _build_api_url(url, query)
        self.assertEqual(
            built_url, "https://api.test.com/v1/endpoint?param1=value1&param2=value2"
        )

    @patch("ten99policy.api_requestor.APIRequestor.request_raw")
    def test_request(self, mock_request_raw):
        mock_request_raw.return_value = (
            json.dumps({"result": "success"}),
            200,
            {},
            self.api_key,
        )
        resp, api_key = self.requestor.request("get", "/v1/test")
        self.assertEqual(resp.data, {"result": "success"})
        self.assertEqual(api_key, self.api_key)

    @patch("ten99policy.api_requestor.APIRequestor.request_raw")
    def test_request_stream(self, mock_request_raw):
        mock_stream = Mock()
        mock_request_raw.return_value = (mock_stream, 200, {}, self.api_key)
        resp, api_key = self.requestor.request_stream("get", "/v1/test")
        self.assertEqual(resp.io, mock_stream)
        self.assertEqual(api_key, self.api_key)

    def test_handle_error_response(self):
        with self.assertRaises(error.InvalidRequestError):
            self.requestor.handle_error_response(
                '{"message": "Invalid request", "error_code": "invalid_request"}',
                400,
                {"message": "Invalid request", "error_code": "invalid_request"},
                {},
            )

    def test_request_headers(self):
        headers = self.requestor.request_headers(self.api_key, "post")
        self.assertIn("Authorization", headers)
        self.assertIn("Content-Type", headers)
        self.assertIn("Idempotency-Key", headers)

    @patch("ten99policy.http_client.RequestsClient.request_with_retries")
    def test_request_raw(self, mock_request_with_retries):
        mock_request_with_retries.return_value = (
            json.dumps({"result": "success"}),
            200,
            {},
        )
        requestor = APIRequestor(key=self.api_key, api_base=self.api_base)
        rcontent, rcode, rheaders, api_key = requestor.request_raw("get", "/v1/test")
        self.assertEqual(json.loads(rcontent), {"result": "success"})
        self.assertEqual(rcode, 200)
        self.assertEqual(api_key, self.api_key)
        mock_request_with_retries.assert_called_once()

    def test_interpret_response_success(self):
        resp = self.requestor.interpret_response(
            json.dumps({"result": "success"}), 200, {}
        )
        self.assertEqual(resp.data, {"result": "success"})

    def test_interpret_response_error(self):
        with self.assertRaises(error.APIError):
            self.requestor.interpret_response(json.dumps({"message": "Error"}), 400, {})

    @patch("ten99policy.api_requestor.Ten99PolicyStreamResponse")
    def test_interpret_streaming_response_success(self, mock_stream_response):
        mock_stream = Mock()
        self.requestor.interpret_streaming_response(mock_stream, 200, {})
        mock_stream_response.assert_called_once_with(mock_stream, 200, {})

    def test_interpret_streaming_response_error(self):
        mock_stream = Mock()
        mock_stream.read.return_value = json.dumps({"message": "Error"}).encode()
        with self.assertRaises(error.APIError):
            self.requestor.interpret_streaming_response(mock_stream, 400, {})


if __name__ == "__main__":
    unittest.main()



# tests/test_util.py
import io
import unittest
from unittest.mock import patch
from ten99policy.util import (
    utf8,
    is_appengine_dev,
    log_debug,
    log_info,
    _test_or_live_environment,
    logfmt,
    secure_compare,
    convert_to_ten99policy_object,
    convert_to_dict,
    populate_headers,
    merge_dicts,
    class_method_variant,
)
import os


# Test with a Ten99PolicyResponse object
class TestUtil(unittest.TestCase):

    def test_utf8(self):
        self.assertEqual(utf8("test"), "test")
        self.assertEqual(utf8("test"), "test")
        self.assertEqual(utf8(b"test"), b"test")

    @patch.dict(os.environ, {"APPENGINE_RUNTIME": "true", "SERVER_SOFTWARE": "Dev"})
    def test_is_appengine_dev(self):
        self.assertTrue(is_appengine_dev())

    @patch.dict(os.environ, {}, clear=True)
    def test_is_not_appengine_dev(self):
        self.assertFalse(is_appengine_dev())

    @patch("ten99policy.util._console_log_level", return_value="debug")
    @patch("ten99policy.util.logger")
    def test_log_debug(self, mock_logger, mock_console_log_level):
        with patch("sys.stderr", new_callable=io.StringIO) as mock_stderr:
            log_debug("test message", key="value")
            self.assertIn("test message", mock_stderr.getvalue())
        mock_logger.debug.assert_called_once()

    @patch("ten99policy.util._console_log_level", return_value="info")
    @patch("ten99policy.util.logger")
    def test_log_info(self, mock_logger, mock_console_log_level):
        with patch("sys.stderr", new_callable=io.StringIO) as mock_stderr:
            log_info("test message", key="value")
            self.assertIn("test message", mock_stderr.getvalue())
        mock_logger.info.assert_called_once()

    @patch("ten99policy.api_key", "sk_test_123")
    def test_test_or_live_environment(self):
        self.assertEqual(_test_or_live_environment(), "test")

    def test_logfmt(self):
        props = {"key": "value", "another_key": "another value"}
        formatted = logfmt(props)
        self.assertIn("key=value", formatted)
        self.assertIn("another_key='another value'", formatted)

    def test_secure_compare(self):
        self.assertTrue(secure_compare("test", "test"))
        self.assertFalse(secure_compare("test", "fail"))

    @patch("ten99policy.ten99policy_object.Ten99PolicyObject")
    def test_convert_to_ten99policy_object(self, MockTen99PolicyObject):
        # Set up the mock to return "converted" when construct_from is called
        MockTen99PolicyObject.construct_from.return_value = "converted"

        # Create an instance of the mocked object
        resp = (
            MockTen99PolicyObject.construct_from()
        )  # Call the mock method to get the expected return value

        # Ensure the function is called with the correct parameters
        result = convert_to_ten99policy_object(resp, "api_key")

        self.assertEqual(result, "converted")

    def test_convert_to_dict(self):
        obj = {"key": "value", "nested": {"nested_key": "nested_value"}}
        result = convert_to_dict(obj)
        self.assertEqual(result, obj)

    def test_populate_headers(self):
        self.assertEqual(populate_headers("key"), {"Ten99Policy-Idempotent-Key": "key"})
        self.assertIsNone(populate_headers(None))

    def test_merge_dicts(self):
        x = {"key1": "value1"}
        y = {"key2": "value2"}
        self.assertEqual(merge_dicts(x, y), {"key1": "value1", "key2": "value2"})

    def test_class_method_variant(self):
        class TestClass:
            @class_method_variant("class_method")
            def instance_method(self):
                return "instance_method"

            @classmethod
            def class_method(cls):
                return "class_method"

        instance = TestClass()
        self.assertEqual(instance.instance_method(), "instance_method")
        self.assertEqual(TestClass.instance_method(), "class_method")


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function

from ten99policy import util
from ten99policy.api_resources.abstract.api_resource import APIResource
from ten99policy.six.moves.urllib.parse import quote_plus


class DeletableAPIResource(APIResource):
    @classmethod
    def _cls_delete(cls, sid, **params):
        url = "%s/%s" % (cls.class_url(), quote_plus(util.utf8(sid)))
        return cls._static_request("delete", url, **params)

    @util.class_method_variant("_cls_delete")
    def delete(self, **params):
        self.refresh_from(self.request("delete", self.instance_url(), params))
        return self



from __future__ import absolute_import, division, print_function

from ten99policy import api_requestor, error, util, six
from ten99policy.ten99policy_object import Ten99PolicyObject
from ten99policy.six.moves.urllib.parse import quote_plus


class APIResource(Ten99PolicyObject):
    @classmethod
    def retrieve(cls, id, api_key=None, **params):
        instance = cls(id, api_key, **params)
        instance.refresh()
        return instance

    def refresh(self):
        self.refresh_from(self.request("get", self.instance_url()))
        return self

    @classmethod
    def class_url(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class.  You should perform "
                "actions on its subclasses (e.g. Charge, Customer)"
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        return "/api/v1/%s" % (base,)

    def instance_url(self):
        id = self.get("id")

        if not isinstance(id, six.string_types):
            raise error.InvalidRequestError(
                "Could not determine which URL to request: %s instance "
                "has invalid ID: %r, %s. ID should be of type `str` (or"
                " `unicode`)" % (type(self).__name__, id, type(id)),
                "id",
            )

        id = util.utf8(id)
        base = self.class_url()
        extn = quote_plus(id)

        return "%s/%s" % (base, extn)

    # The `method_` and `url_` arguments are suffixed with an underscore to
    # avoid conflicting with actual request parameters in `params`.
    @classmethod
    def _static_request(
        cls,
        method_,
        url_,
        api_key=None,
        idempotency_key=None,
        ten99policy_version=None,
        ten99policy_environment=None,
        **params
    ):
        requestor = api_requestor.APIRequestor(
            api_key, api_version=ten99policy_version, environment=ten99policy_environment,
        )
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request(method_, url_, params, headers)
        return util.convert_to_ten99policy_object(
            response, api_key, ten99policy_version, ten99policy_environment
        )

    # The `method_` and `url_` arguments are suffixed with an underscore to
    # avoid conflicting with actual request parameters in `params`.
    @classmethod
    def _static_request_stream(
        cls,
        method_,
        url_,
        api_key=None,
        idempotency_key=None,
        ten99policy_version=None,
        ten99policy_environment=None,
        **params
    ):
        requestor = api_requestor.APIRequestor(
            api_key, api_version=ten99policy_version, environment=ten99policy_environment,
        )
        headers = util.populate_headers(idempotency_key)
        response, _ = requestor.request_stream(method_, url_, params, headers)
        return response



import unittest
import sys
import types
import io

import ten99policy.six as six


class TestSix(unittest.TestCase):

    def test_python_version(self):
        self.assertIn(six.PY2, [True, False])
        self.assertIn(six.PY3, [True, False])
        self.assertNotEqual(six.PY2, six.PY3)

    def test_string_types(self):
        if six.PY3:
            self.assertEqual(six.string_types, (str,))
        else:
            self.assertEqual(six.string_types, (basestring,))

    def test_integer_types(self):
        if six.PY3:
            self.assertEqual(six.integer_types, (int,))
        else:
            self.assertEqual(six.integer_types, (int, long))

    def test_class_types(self):
        if six.PY3:
            self.assertEqual(six.class_types, (type,))
        else:
            self.assertEqual(six.class_types, (type, types.ClassType))

    def test_text_type(self):
        if six.PY3:
            self.assertEqual(six.text_type, str)
        else:
            self.assertEqual(six.text_type, unicode)

    def test_binary_type(self):
        if six.PY3:
            self.assertEqual(six.binary_type, bytes)
        else:
            self.assertEqual(six.binary_type, str)

    def test_maxsize(self):
        self.assertEqual(six.MAXSIZE, sys.maxsize)

    def test_b(self):
        self.assertEqual(six.b("foo"), b"foo")

    def test_u(self):
        self.assertEqual(six.u("foo"), "foo")

    def test_unichr(self):
        self.assertEqual(six.unichr(97), "a")

    def test_int2byte(self):
        self.assertEqual(six.int2byte(65), b"A")

    def test_byte2int(self):
        self.assertEqual(six.byte2int(b"A"), 65)

    def test_indexbytes(self):
        self.assertEqual(six.indexbytes(b"ABC", 1), 66)

    def test_iterbytes(self):
        self.assertEqual(list(six.iterbytes(b"ABC")), [65, 66, 67])

    def test_StringIO(self):
        self.assertTrue(isinstance(six.StringIO(), io.StringIO))

    def test_BytesIO(self):
        self.assertTrue(isinstance(six.BytesIO(), io.BytesIO))

    def test_exec_(self):
        namespace = {}
        six.exec_("a = 1", namespace)
        self.assertEqual(namespace["a"], 1)

    def test_reraise(self):
        try:
            try:
                raise ValueError("test")
            except ValueError:
                six.reraise(*sys.exc_info())
        except ValueError as e:
            self.assertEqual(str(e), "test")

    def test_raise_from(self):
        try:
            six.raise_from(ValueError("test"), TypeError("original"))
        except ValueError as e:
            self.assertEqual(str(e), "test")
            if six.PY3:
                self.assertIsInstance(e.__cause__, TypeError)

    def test_print_(self):
        buf = six.StringIO()
        six.print_("Hello", "world!", file=buf)
        self.assertEqual(buf.getvalue().strip(), "Hello world!")

    def test_with_metaclass(self):
        class Meta(type):
            pass

        class Base(object):
            pass

        class MyClass(six.with_metaclass(Meta, Base)):
            pass

        self.assertTrue(isinstance(MyClass, Meta))
        self.assertTrue(issubclass(MyClass, Base))

    def test_add_metaclass(self):
        class Meta(type):
            pass

        @six.add_metaclass(Meta)
        class MyClass(object):
            pass

        self.assertTrue(isinstance(MyClass, Meta))

    def test_python_2_unicode_compatible(self):
        @six.python_2_unicode_compatible
        class MyClass(object):
            def __str__(self):
                return "unicode"

        self.assertEqual(str(MyClass()), "unicode")
        if six.PY2:
            self.assertTrue(hasattr(MyClass, "__unicode__"))

    def test_wraps(self):
        def decorator(func):
            @six.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        @decorator
        def example():
            """Docstring"""
            pass

        self.assertEqual(example.__name__, "example")
        self.assertEqual(example.__doc__, "Docstring")

    def test_iterkeys(self):
        d = {"a": 1, "b": 2}
        self.assertEqual(list(six.iterkeys(d)), ["a", "b"])

    def test_itervalues(self):
        d = {"a": 1, "b": 2}
        self.assertEqual(list(six.itervalues(d)), [1, 2])

    def test_iteritems(self):
        d = {"a": 1, "b": 2}
        self.assertEqual(list(six.iteritems(d)), [("a", 1), ("b", 2)])

    def test_callable(self):
        def func():
            pass

        self.assertTrue(six.callable(func))
        self.assertFalse(six.callable(1))


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch, Mock
import datetime
import json

from ten99policy.ten99policy_object import Ten99PolicyObject
from ten99policy import util, api_requestor


class TestTen99PolicyObject(unittest.TestCase):

    def setUp(self):
        self.obj = Ten99PolicyObject(
            id="obj_123",
            api_key="sk_test_123",
            ten99policy_version="2023-03-01",
            ten99policy_environment="test",
        )

    def test_init(self):
        self.assertEqual(self.obj.id, "obj_123")
        self.assertEqual(self.obj.api_key, "sk_test_123")
        self.assertEqual(self.obj.ten99policy_version, "2023-03-01")
        self.assertEqual(self.obj.ten99policy_environment, "test")

    def test_update(self):
        self.obj.update({"foo": "bar"})
        self.assertEqual(self.obj["foo"], "bar")
        self.assertIn("foo", self.obj._unsaved_values)

    def test_setattr(self):
        self.obj.new_attr = "value"
        self.assertEqual(self.obj["new_attr"], "value")
        self.assertIn("new_attr", self.obj._unsaved_values)

    def test_getattr(self):
        self.obj["get_attr"] = "test"
        self.assertEqual(self.obj.get_attr, "test")

        with self.assertRaises(AttributeError):
            self.obj.nonexistent_attr

    def test_delattr(self):
        self.obj["del_attr"] = "test"
        del self.obj.del_attr
        self.assertNotIn("del_attr", self.obj)

    def test_setitem_empty_string(self):
        with self.assertRaises(ValueError):
            self.obj["empty"] = ""

    def test_getitem_transient(self):
        self.obj._transient_values.add("transient")
        with self.assertRaises(KeyError):
            self.obj["transient"]

    def test_serialize(self):
        self.obj["serialize_me"] = "value"
        serialized = self.obj.serialize({})
        self.assertEqual(serialized["serialize_me"], "value")

    @patch("ten99policy.util.convert_to_ten99policy_object")
    def test_refresh_from(self, mock_convert):
        mock_convert.return_value = "converted"
        self.obj.refresh_from({"new_key": "new_value"})
        self.assertEqual(self.obj["new_key"], "converted")

    @patch("ten99policy.api_requestor.APIRequestor")
    def test_request(self, mock_requestor):
        mock_instance = Mock()
        mock_requestor.return_value = mock_instance
        mock_instance.request.return_value = ({"id": "resp_123"}, "sk_test_123")

        response = self.obj.request("get", "/v1/endpoint")

        mock_requestor.assert_called_once_with(
            key="sk_test_123",
            api_base=None,
            api_version="2023-03-01",
            environment="test",
        )
        # Update the expected params to be an empty dict instead of None
        mock_instance.request.assert_called_once_with("get", "/v1/endpoint", {}, None)
        self.assertIsNotNone(response)

    def test_to_dict_recursive(self):
        nested_obj = Ten99PolicyObject(id="nested_123")
        self.obj["nested"] = nested_obj
        self.obj["list"] = [1, nested_obj]

        result = self.obj.to_dict_recursive()

        self.assertIsInstance(result["nested"], dict)
        self.assertEqual(result["nested"]["id"], "nested_123")
        self.assertIsInstance(result["list"][1], dict)

    def test_repr_json_encoder(self):
        encoder = Ten99PolicyObject.ReprJSONEncoder()
        # Use UTC time to avoid time zone issues
        dt = datetime.datetime(2023, 3, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)
        encoded = encoder.encode({"date": dt})

        # Parse the encoded JSON
        decoded = json.loads(encoded)

        # Check if the 'date' key exists and its value is a number (timestamp)
        self.assertIn("date", decoded)
        self.assertTrue(isinstance(decoded["date"], (int, float)))

        # Convert the timestamp back to a datetime object (assuming UTC)
        encoded_dt = datetime.datetime.fromtimestamp(
            decoded["date"], tz=datetime.timezone.utc
        )

        # Check if the encoded datetime is exactly the same as the original
        self.assertEqual(encoded_dt, dt)

    def test_construct_from(self):
        values = {"id": "const_123", "key": "value"}
        obj = Ten99PolicyObject.construct_from(values, "sk_test_123")
        self.assertEqual(obj.id, "const_123")
        self.assertEqual(obj["key"], "value")

    def test_pickling(self):
        import pickle

        pickled = pickle.dumps(self.obj)
        unpickled = pickle.loads(pickled)
        self.assertEqual(self.obj.id, unpickled.id)
        self.assertEqual(self.obj.api_key, unpickled.api_key)

    def test_serialize_empty_string(self):
        self.obj["empty"] = None
        serialized = self.obj.serialize({})
        self.assertEqual(serialized["empty"], "")

    def test_dirty_json_parse(self):
        values = {"key": "{'nested': 'value'}"}
        obj = Ten99PolicyObject.construct_from(values, "sk_test_123")
        self.assertEqual(obj["key"], {"nested": "value"})

    def test_copy(self):
        self.obj["copy_me"] = "value"
        copied = self.obj.__copy__()
        self.assertEqual(copied["copy_me"], "value")
        self.assertIsNot(copied, self.obj)

    def test_deepcopy(self):
        nested = {"nested": "value"}
        self.obj["deep_copy"] = nested
        deep_copied = self.obj.__deepcopy__({})
        self.assertEqual(deep_copied["deep_copy"], nested)
        self.assertIsNot(deep_copied["deep_copy"], nested)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function

from ten99policy.ten99policy_object import Ten99PolicyObject


class ErrorObject(Ten99PolicyObject):
    def refresh_from(
        self,
        values,
        api_key=None,
        partial=False,
        ten99policy_version=None,
        ten99policy_environment=None,
        last_response=None,
    ):
        return super(ErrorObject, self).refresh_from(
            values,
            api_key,
            partial,
            ten99policy_version,
            ten99policy_environment,
            last_response,
        )



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from ten99policy import api_resources


OBJECT_CLASSES = {
    # data structures
    api_resources.ListObject.OBJECT_NAME: api_resources.ListObject,
    api_resources.Contractors.OBJECT_NAME: api_resources.Contractors,
    api_resources.Entities.OBJECT_NAME: api_resources.Entities,
    api_resources.Jobs.OBJECT_NAME: api_resources.Jobs,
    api_resources.Policies.OBJECT_NAME: api_resources.Policies,
    api_resources.Quotes.OBJECT_NAME: api_resources.Quotes,
}
